import anvil.server
import anvil._threaded_server
from random import SystemRandom

_components = {}


def _mkstr(length):
    sr = SystemRandom()
    return "".join((sr.choice("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890") for i in range(length)))


def _real_component_to_pdf(options, component_args, component_kwargs):
    key = _mkstr(20)
    _components[key] = (component_args, component_kwargs)
    try:
        return anvil.server.call("anvil.private.pdf.do_print", [anvil._threaded_server.call_info.call_id, key], options)
    finally:
        del _components[key]


@anvil.server.callable('anvil.private.pdf.component_to_pdf')
def _component_to_pdf(options, component_args, component_kwargs):
    if not anvil.server.context.remote_caller.is_trusted:
        raise Exception("Can only generate PDFs from server code")
    return _real_component_to_pdf(options, component_args, component_kwargs)

#!defFunction(anvil.pdf,%anvil.Media instance,form_name,*args,**kwargs)!2: "Render an Anvil form to PDF. Pass the name of the form you want to render, plus any arguments you want to pass to its constructor.\n\nReturns a PDF as an Anvil Media object." ["render_form"]
def render_form(*args, **kwargs):
    return _real_component_to_pdf({}, args, kwargs)


form_to_pdf = render_form


class PDFRenderer(object):
    #!defMethod(_,[filename="print.pdf"],[landscape=False],[margins=],[page_size="letter"],[quality="default"],[scale=1.0])!2: "Configure options for PDF rendering. Returns an object with a render_form() method." ["__init__"]
    def __init__(self, **kwargs):
        self.options = kwargs

    #!defMethod(anvil.Media instance,form_name,*args,**kwargs)!2: "Render an Anvil form to PDF. Pass the name of the form you want to render, plus any arguments you want to pass to its constructor.\n\nReturns a PDF as an Anvil Media object." ["render_form"]
    def render_form(self, *args, **kwargs):
        return _real_component_to_pdf(self.options, args, kwargs)
#!defClass(anvil.pdf,%PDFRenderer)!0:


@anvil.server.callable('anvil.private.pdf.get_component')
def _get_component(pk):
    _, print_key = pk
    return _components[print_key]



# This file is automatically generated by gendoc

from . import server, _server
@server.portable_class('anvil.Component')
class Component(_server.SerializeWithIdentity):
    def __init__(self,__ignore_property_exceptions=False,**props):
        self.__dict__.update(props)
    def __deserialize__(self, data, global_data):
        self.__dict__.update(data)
    def __serialize_once__(self, global_data):
        return self.__dict__

@server.portable_class('anvil.Container')
class Container(Component):
    def __init__(self,__ignore_property_exceptions=False,**props):
        self.__dict__['$_components'] = []
        self.__dict__.update(props)
    def add_component(self, c, **lp):
        self.__dict__['$_components'].append((c, lp))

@server.portable_class('anvil.ComponentTag')
class ComponentTag(object):
    pass

@server.portable_class('anvil.Button')
class Button(Component):
    pass

@server.portable_class('anvil.Canvas')
class Canvas(Component):
    pass

@server.portable_class('anvil.CheckBox')
class CheckBox(Component):
    pass

@server.portable_class('anvil.ColumnPanel')
class ColumnPanel(Container):
    pass

@server.portable_class('anvil.DataGrid')
class DataGrid(Container):
    pass

@server.portable_class('anvil.DataRowPanel')
class DataRowPanel(Container):
    pass

@server.portable_class('anvil.DatePicker')
class DatePicker(Component):
    pass

@server.portable_class('anvil.DropDown')
class DropDown(Component):
    pass

@server.portable_class('anvil.FileLoader')
class FileLoader(Component):
    pass

@server.portable_class('anvil.FlowPanel')
class FlowPanel(Container):
    pass

@server.portable_class('anvil.GoogleMap')
class GoogleMap(Container):
    pass

@server.portable_class('anvil.GridPanel')
class GridPanel(Container):
    pass

@server.portable_class('anvil.HtmlTemplate')
class HtmlTemplate(Container):
    pass

@server.portable_class('anvil.Image')
class Image(Component):
    pass

@server.portable_class('anvil.Label')
class Label(Component):
    pass

@server.portable_class('anvil.LinearPanel')
class LinearPanel(Container):
    pass

@server.portable_class('anvil.Link')
class Link(Container):
    pass

@server.portable_class('anvil.Plot')
class Plot(Component):
    pass

@server.portable_class('anvil.RadioButton')
class RadioButton(Component):
    pass

@server.portable_class('anvil.RepeatingPanel')
class RepeatingPanel(Component):
    pass

@server.portable_class('anvil.RichText')
class RichText(Container):
    pass

@server.portable_class('anvil.Spacer')
class Spacer(Component):
    pass

@server.portable_class('anvil.TextArea')
class TextArea(Component):
    pass

@server.portable_class('anvil.TextBox')
class TextBox(Component):
    pass

@server.portable_class('anvil.Timer')
class Timer(Component):
    pass

@server.portable_class('anvil.XYPanel')
class XYPanel(Container):
    pass

@server.portable_class('anvil.YouTubeVideo')
class YouTubeVideo(Component):
    pass




import matplotlib.pyplot as plt
import anvil
import io


#!defFunction(anvil.mpl_util,anvil.Media instance,[dpi=],[facecolor=],[edgecolor=],[format=],[transparent=],[frameon=],[bbox_inches=],[pad_inches=],[filename=])!2: "Return the current Matplotlib figure as an PNG image. Returns an Anvil Media object that can be displayed in Image components.\n\nOptional arguments have the same meaning as for 'savefig()'" ["plot_image"]
def plot_image(format='png', transparent=True, **kwargs):
  with io.BytesIO() as buf:
    plt.savefig(buf, format=format, transparent=transparent, **kwargs)
    buf.seek(0)    
    return anvil.BlobMedia('image/png', buf.read(), name=kwargs.get("filename", "plot.png"))



import anvil.server
from anvil import *
from ..exceptions import AuthenticationFailed
from ..config import get_client_config

#!defFunction(anvil.users,_,email_address)!2: "Send a two-factor authentication reset email to the specified user." ["send_mfa_reset_email"]
def send_mfa_reset_email(email):
    anvil.server.call("anvil.private.users.send_mfa_reset_email", email)
if is_server_side():
    pass
else:
    from . import webauthn

    #!defFunction(anvil.users.mfa,_,email_address)!2: "Generate a WebAuthn challenge that can be used to register a new hardware token for two-factor authentication." ["create_fido_mfa_method"]
    def create_fido_mfa_method(email):
        opts = anvil.server.call("anvil.private.users.begin_fido_attestation", email)
        result = webauthn.create(opts)
        if result:
            return anvil.server.call("anvil.private.users.validate_fido_attestation", result)

    #!defFunction(anvil.users.mfa,_,email_address,password)!2: "Generate a WebAuthn challenge that the given user can use to log in with a previously registered hardware token." ["get_fido_mfa_login"]
    def get_fido_mfa_login(email, password):
        opts = anvil.server.call("anvil.private.users.begin_fido_assertion", email, password)
        result = webauthn.get(opts)
        if result:
            return {"type": 'fido', "result": result}

    #!defFunction(anvil.users.mfa,_,code)!2: "Get an MFA login object representing a TOTP login code. This can be passed to the login_with_email function as the mfa argument." ["get_totp_mfa_login"]
    def get_totp_mfa_login(code):
        return {"type": "totp", "code": code}

    #!defFunction(anvil.users.mfa,_,email_address)!2: "Generate a TOTP secret that can be added as two-factor authentication for the current user." ["generate_totp_secret"]
    def generate_totp_secret(email):
        return anvil.server.call("anvil.private.users.totp.generate_secret", email)

    #!defFunction(anvil.users.mfa,_,mfa_method,code)!2: "Validate the given TOTP code against the given MFA method from a User row." ["validate_totp_code"]
    def validate_totp_code(mfa_method, code):
        return anvil.server.call("anvil.private.users.totp.validate_code", mfa_method, code)

    #!defFunction(anvil.users.mfa,_,password,mfa_method,[clear_existing=False])!2: "Add an MFA method to the current user by passing the user's password and the mfa method, optionally clearing all existing methods." ["add_mfa_method"]
    def add_mfa_method(password, method, clear_existing=False):
        return anvil.server.call("anvil.private.users.add_mfa_method", password, method, clear_existing)

    #!defFunction(anvil.users.mfa,_,email_address,password)!2: "Get the available MFA types for the given user by passing their email and password." ["get_available_mfa_types"]
    def get_available_mfa_types(email, password):
        return anvil.server.call("anvil.private.users.get_available_mfa_types", email, password)

    def _configure_mfa(email, mfa_error, require_password, allow_cancel, confirm_button_text):
        totp_config = generate_totp_secret(email)
        totp_secret = totp_config['secret']
        qr_code = totp_config['qr_code']
        mfa_methods = {
            'totp': totp_config['mfa_method'],
        }

        while True:

            mfa_panel = LinearPanel()
            mfa_panel.add_component(Label(text="This app requires 2-factor authentication to log in."))

            def signup_with_fido(**e):
                result = create_fido_mfa_method(email)
                if result:
                    mfa_methods['fido'] = result
                    mfa_panel.raise_event("x-close-alert", value='fido')

            if webauthn.is_webauthn_available():
                fido_link = Link(text="Use hardware token",icon="fa:lock")
                fido_link.set_event_handler("click", signup_with_fido)
                mfa_panel.add_component(fido_link)
            else:
                mfa_panel.add_component(Label(text="Hardware token unavailable", icon="fa:lock", tooltip="Is this page running in an iframe or an unsupported browser?"))

            mfa_panel.add_component(Label(text="Alternatively, scan this QR-code with your Authenticator app and then enter the current code below to continue."))
            mfa_panel.add_component(Image(source=qr_code, display_mode="fill_width"))

            totp_box = TextBox(placeholder="Enter 6-digit code", align="center", font="monospace")
            totp_box.set_event_handler("pressed_enter", lambda **e: mfa_panel.raise_event("x-close-alert", value='totp'))

            password_box = TextBox(placeholder="Current password", align="center", hide_text=True)
            password_box.set_event_handler("show", lambda **e: password_box.focus())
            password_box.set_event_handler("pressed_enter", lambda **e: totp_box.focus())

            if require_password:
                mfa_panel.add_component(password_box)
                password_box.set_event_handler("show", lambda **e: password_box.focus())
            else:
                totp_box.set_event_handler("show", lambda **e: totp_box.focus())

            mfa_panel.add_component(totp_box)



            if mfa_error:
                mfa_panel.add_component(Label(foreground="red", bold=True, spacing_below="none", text=mfa_error))

            maybe_cancel_button = [("Cancel", None)] if allow_cancel else []
            mfa = alert(mfa_panel, title="2-Factor Authentication", buttons=[(confirm_button_text, 'totp', 'success')] + maybe_cancel_button, dismissible=bool(allow_cancel))
            if mfa == 'totp': 
                if validate_totp_code(mfa_methods['totp'], totp_box.text):
                    return mfa_methods['totp'], password_box.text
                else:
                    mfa_error = "Incorrect code entered. Please try again."
                    continue
            elif mfa == 'fido':
                return mfa_methods['fido'], password_box.text
            else:
                return None, ""


    #!defFunction(anvil.users.mfa,!,email_address, password)!2: "Display a form to collect two-factor authentication credentials from the user currently logging in by passing the function their email and password." ["mfa_login_with_form"]
    def mfa_login_with_form(email, password):
        mfa_panel = LinearPanel()

        mfa_types = get_available_mfa_types(email, password)

        maybe_login_button = []
        
        if 'totp' in mfa_types:
            totp_box = TextBox(placeholder="Enter 6-digit code", align="center", font="monospace")
            totp_box.set_event_handler("pressed_enter", lambda **e: mfa_panel.raise_event('x-close-alert', value=get_totp_mfa_login(totp_box.text)))
            totp_box.set_event_handler("show", lambda **e: totp_box.focus())
            mfa_panel.add_component(Label(text="Please enter the 2-factor authentication code from your Authenticator app"))
            mfa_panel.add_component(totp_box)
            maybe_login_button = [("Log In", 'totp', 'success')]

        if 'fido' in mfa_types:
            def use_fido(**e):
                val = get_fido_mfa_login(email, password)
                if val:
                    mfa_panel.raise_event("x-close-alert", value=val)

            
            if webauthn.is_webauthn_available():
                fido_link = Link(text="Use hardware token", icon="fa:lock")
                fido_link.set_event_handler("click", use_fido)
                mfa_panel.add_component(fido_link)
            else:
                mfa_panel.add_component(Label(text="Hardware token unavailable", icon="fa:lock", tooltip="Is this page running in an iframe or an unsupported browser?"))


        if not mfa_types:
            mfa_panel.add_component(Label(text="No authentication methods available."))

        if not mfa_types or get_client_config().get("allow_mfa_email_reset", False):
            mfa_reset_link = Link(text="Reset 2-factor authentication by email")
            mfa_reset_link.set_event_handler('click', lambda **e: mfa_panel.raise_event('x-close-alert', value='reset_mfa'))
            mfa_panel.add_component(mfa_reset_link)

        r = alert(mfa_panel, title="2-Factor Authentication", buttons=maybe_login_button + [("Cancel", None, 'default')], dismissible=False)

        if r == 'totp':
            return get_totp_mfa_login(totp_box.text)
        else:
            return r


    #!defFunction(anvil.users.mfa,!,[allow_cancel=False])!2: "Display a form for the user to configure 2-factor authentication.\n\nallow_cancel: if True, the signup form has a Cancel button that the user can use to dismiss the form." ["configure_mfa_with_form"]
    def configure_mfa_with_form(allow_cancel=False):

        error = None
        while True:
            mfa_method, password = _configure_mfa(None, error, True, allow_cancel, "Save")

            if mfa_method:
                try:
                    add_mfa_method(password, mfa_method)
                    alert("Your two-factor authentication configuration has been reset.")
                    return True
                except AuthenticationFailed as e:
                    error = e.args[0]
                except Exception as e:
                    error = str(e)
            else:
                return None



import anvil.google.auth
import anvil.regex
import anvil.server as rpc

add_missing_fields = False

def index_to_col(idx):
    idx -= 1
    if idx < 26:
        return chr(65 + (idx % 26))
    else:
        idx -= 26
        return chr(65 + (idx // 26)) + chr(65 + (idx % 26))

class ApiItem(object):

    def __getitem__(self, item_name):

        if item_name in self._obj:
            return self._obj[item_name]
        else:
            raise KeyError(item_name)

    def __getattr__(self, attr_name):
        if attr_name in self._obj:
            return self._obj[attr_name]
        elif attr_name == "id":
            return self._obj["id"]
        else:
            raise AttributeError("Could not find attribute: %s" % attr_name)

    def __init__(self, obj, other):
        self._obj = obj
        self._other = other

@anvil.server.serializable_type
class Cell(ApiItem):

    #!defAttr()!1: {name:"row",type:"number",description:"This cell's row index (starting from 1)"}
    #!defAttr()!1: {name:"col",type:"number",description:"This cell's column index (starting from 1)"}
    #!defAttr()!1: {name:"value",type:"string",description:"The value in this cell"}
    def __getattr__(self, name):
        if name == "input_value":
            # This used to be different. In v4 we just return the value, because it would take two API calls to get the input_value too.
            return self["value"]
        else:
            return ApiItem.__getattr__(self, name)

    def __setattr__(self, name, value):
        if name == "value" or name == "inputValue":

            rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"], value)

            self._obj['value'] = value

            return value
        else:
            object.__setattr__(self, name, value)

    def __repr__(self):
        return "<Google Worksheet Cell: %s>" % self.value
#!defClass(anvil.google.sheets,#Cell)!:


def old_field_transform(new_field_name):
    return anvil.regex.replace(new_field_name, "[^A-Za-z0-9\\-]", "").lower()

@anvil.server.serializable_type
class Row(ApiItem):
    def __getitem__(self, key):
        return self._other["data"][self._other['fields_old'].get(key, key)]

    def __setitem__(self, key, value):
        key = self._other['fields_old'].get(key, key)
        if key in self._other['data']:
            rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"].narrow([index_to_col(self._other['fields'].index(key)+1)]), value)
        else:
            raise KeyError(key)

    #!defMethod(_)!2: "Delete this row from the worksheet. (This will cause data in subsequent rows to shift up)" ["delete"]
    def delete(self):
        rpc.call("anvil.private.google.sheets.v4.delete_row", self._other["capability"])

    def __iter__(self):
        return iter(self._other['data'])

    def __repr__(self):
        return "<Google Worksheet Row: %s>" % self._other['data']
#!defClass(anvil.google.sheets,#Row)!:


class RowIterator(object):

    def __init__(self, cap, query_dict):
        self.cap = cap

        self.fields, rows, self.next_start, self.done = rpc.call("anvil.private.google.sheets.v4.list_initial_rows", cap, query_dict)
        self.query_list = [query_dict.get(f) for f in self.fields]
        self.fields_old = {old_field_transform(f): f for f in self.fields}
        self.iterator = iter(rows)

    def _load_next_page(self):
        rows, self.next_start, self.done = rpc.call("anvil.private.google.sheets.v4.list_more_rows", self.cap, self.query_list, self.next_start)
        self.iterator = iter(rows)

    def __next__(self):
        while True:
            try:
                r = next(self.iterator)
                return Row({}, {
                    "data": {self.fields[i]: (r['data'][i] if i < len(r['data']) else "") for i in range(len(self.fields))},
                    "capability": self.cap.narrow([r['row']]),
                    "fields": self.fields,
                    "fields_old": self.fields_old
                })
            except StopIteration:
                if self.done:
                    raise
                else:
                    self._load_next_page()

    next = __next__


class RowList(object):

    def __init__(self, cap, query_dict):
        self.cap = cap
        self.query_dict = query_dict
        self.it = RowIterator(cap, query_dict)

    def __iter__(self):
        if not self.it:
            return RowIterator(self.cap, self.query_dict)
        try:
            return self.it
        finally:
            self.it = None


@anvil.server.serializable_type
class Worksheet(ApiItem):

    #!defAttr()!1: {name:"title",type:"string",description:"The title of this worksheet"}
    #!defAttr()!1: {name:"row_count",type:"number",description:"The number of rows in this worksheet"}
    #!defAttr()!1: {name:"column_count",type:"number",description:"The number of columns in this worksheet"}
    #!defAttr()!1: {name:"fields",type:"list",description:"The fields in this worksheet (ie the column headers, or the values in the first row)"}
    #!defAttr()!1: {name:"rows",pyType:"list(anvil.google.sheets.Row instance)",description:"The rows in this worksheet (excluding the header)"}
    #!defAttr()!1: {name:"cells",pyType:"list(anvil.google.sheets.Cell instance)",description:"A list of all the cells in this worksheet"}
    def __getattr__(self, name):
        if name == "fields":
            return self.get_fields()
        elif name == "rows":
            return list(self.list_rows())
        elif name == "cells":
            return list(self.list_cells())
        else:
            return ApiItem.__getattr__(self, name)

    #!setItemType(anvil.google.sheets.Cell instance)!:
    def __getitem__(self, cell_tuple):
        row, col = cell_tuple

        return self.get_cell(row, col)

    def get_fields(self):
        cs = self.column_count

        return [c.value for c in self.list_cells(1,1,1,cs)]

    #!defMethod(list[anvil.google.sheets.Row instance],**query)!2: "List rows in this worksheet, optionally restricting to rows with the specified column values specified as keyword arguments" ["list_rows"]
    def list_rows(self, **query):
        return RowList(self._other['capability'], query)

    #!defMethod(anvil.google.sheets.Row instance,**fields)!2: "Add a row to the end of the worksheet, specifying values for columns as keywords arguments" ["add_row"]
    def add_row(self, **kwargs):

        fields = self.get_fields()
        fields_old = {old_field_transform(f): f for f in fields}

        values = {}

        for [k,v] in kwargs.items():
            key = fields_old.get(k,k)
            values[key] = v

        data = []
        for i in range(len(fields)):
            data.append(values.get(fields[i],""))

        added_row_index = rpc.call("anvil.private.google.sheets.v4.add_row", self._other['capability'], data)

        return Row({}, {
            "data": {fields[i]: data[i] for i in range(len(fields))},
            "capability": self._other["capability"].narrow([added_row_index]),
            "fields": fields,
            "fields_old": fields_old
        })

    #!defMethod(list[anvil.google.sheets.Cell instance],[min_row=],[max_row=],[min_col=],[max_col=])!2: "List cells in the worksheet, optionally specifying a region" ["list_cells"]
    def list_cells(self, min_row=None, max_row=None, min_col=None, max_col=None):

        min_col_str = index_to_col(min_col)
        max_col_str = index_to_col(max_col)

        range = min_col_str + str(min_row) + ":" + max_col_str + str(max_row)

        cells = rpc.call("anvil.private.google.sheets.v4.list_cells", self._other['capability'], range)
        return [Cell(c,{
            "capability": self._other['capability'].narrow([c['row'], index_to_col(c['col'])])
        }) for c in cells]

    #!defMethod(anvil.google.sheets.Cell instance,row,col)!2: "Get a particular cell from the spreadsheet" ["get_cell"]
    def get_cell(self, row, col):
        cells = self.list_cells(row, row, col, col)
        if len(cells) > 0:
            return cells[0]
        return None


    def __repr__(self):
        return "<Google Worksheet: %s>" % self._obj['title']
#!defClass(anvil.google.sheets,#Worksheet)!:


@anvil.server.serializable_type
class Spreadsheet(ApiItem):
    #!setItemType(anvil.google.sheets.Worksheet instance)!:
    def __getitem__(self, name):
      if isinstance(name, str):
        for w in self.list_worksheets():
          if w.title == name:
            return w
        raise KeyError("Spreadsheet contains no worksheet '%s'" % name)
      else:
        return list(self.list_worksheets())[name]

    #!defAttr()!1: {name:"worksheets",pyType:"list(anvil.google.sheets.Worksheet instance)",description:"The worksheets in this spreadsheet."}
    #!defAttr()!1: {name:"title",type:"string",description:"The title of this spreadsheet."}
    #!defAttr()!1: {name:"id",type:"string",description:"The ID of this spreadsheet in Google Drive"}
    def __getattr__(self, name):
        if name in self._obj:
            return self._obj[name]
        elif name == "worksheets":
            return self
        else:
            raise AttributeError("Could not find attribute: %s" % name)

    def __iter__(self):
        return iter(self.list_worksheets())

    def __len__(self):
        return len(list(self.list_worksheets()))

    #!defMethod(list[anvil.google.sheets.Worksheet instance])!2: "Get a list of all worksheets in this spreadsheet" ["list_worksheets"]
    def list_worksheets(self):
        return [Worksheet({
            "title": w['title'],
            "row_count": w['gridProperties']['rowCount'],
            "column_count": w['gridProperties']['columnCount']
        }, {
            "capability": self._other['capability'].narrow([[w['title'], w['sheetId']]])
        }) for w in self._other['worksheets']]


    def __repr__(self):
        return "<Google Sheet: %s>" % self._obj["title"]
#!defClass(anvil.google.sheets,#Spreadsheet)!:



Sheet = Spreadsheet # Alias for backwards-compatibility


def login():
    return anvil.google.auth.login(['https://www.googleapis.com/auth/spreadsheets'])


def get_sheet(id, creds):
    s = Spreadsheet(**rpc.call("anvil.private.google.sheets.v4.get_sheet", id, creds))
    return s


"""
id: drive_sheets
docs_url: /docs/integrations/google/google-drive#google-sheets
title: Google Sheets
description: |
  ```python
  db = app_files.my_sheet
  ```

  Google Sheets make ideal databases for Anvil apps. You can access sheets using Google Drive in the same way you access files.

  ```python
  ws = db["Sheet 1"]
  ```

  Once you have an object representing a Google Sheet, you can look up its individual worksheets. (It is also possible to look up a worksheet by number, eg `db[0]`, or obtain a list with `list(db.worksheets)`.)

  ```python
  print db["Sheet 1"].fields
  # Expect output similar to: '[name, age]'
  ```

  You should set up your Sheet using [Google Drive](https://drive.google.com) so that the column names are in the first row. Once your sheet is set up, you can get a list of the fields in a worksheet using the `fields` attribute.

  Fields will always be in lower case (even if your column headings use capital letters).

  ```python
  # Assuming columns "name" and "age"
  for r in db["Sheet 1"].rows:
    print "%s is %s years old" % (r["name"], r["age"])
  ```

  You can access rows of data like so:

  ```python
  # Convert age to dog years:
  for r in db["Sheet 1"].rows:
    r["age"] = 7 * int(r["age"])
  ```

  You can update them as well as reading them. You can assign numbers or strings,
  but the data you read from a row will always be a string:

  ```python
  ws = db["Sheet 1"]
  row = ws.add_row(name="Bob", age=56)
  ```

  You can add rows to your spreadsheet using the `add_row(...)` method:

  (Note: `add_row()` normally returns the new created row. However, if you
  attempt to insert an empty row, it will return `None`.)

  An Exception will be raised if you try to set a non-existent field.

  ```python
  row.delete()
  ```

  You can delete a row using the `delete()` method.

  Beware: If you delete a row, other rows in the worksheet may shift unpredictably. If you delete a row, you should not re-use any other `row` objects from that worksheet. Call `list_rows()` again to find the rows you want again.

  ```python
  ## Display the calculated value of
  ## the cell in row 1, col 2

  print ws[1,2].value

  ## Display the formula in cell (1,2)

  print ws[1,2].input_value

  ## Set the value of a cell

  ws[3,4].value = "Some text"
  ```

  It is also possible to access individual cells of your spreadsheet.

"""



from __future__ import absolute_import
import sys, imp, importlib

# The downlink may need to coexist with the Uplink (for example in the standalone App Server).
# In these deployments, the downlink's version of the 'anvil' module is shipped as
# anvil_downlink_worker.anvil, and so we do some path gymnastics to load it from there:
old_path = sys.path
sys.path = __path__ + sys.path
import anvil
sys.path = old_path

import anvil.server
from anvil import _server, _serialise, _threaded_server, _form_templating

if sys.version_info[0] < 3:
    from .exec2 import do_exec
else:
    from .exec3 import do_exec

_server._do_call = _threaded_server.do_call
_serialise.holding_reqresps = True # Don't do anything until we've loaded apps


def find(lst, f):
    for item in lst:
        if f(item):
            return item


def find_in_app(app, mod_name):
    return find(app.get('server_modules', []), lambda m: m['name'] == mod_name) or \
           find(app.get('modules', []), lambda m: m['name'] == mod_name)


class ErrorLoadingUserCode(Exception):
    def __init__(self, exc):
        self.exc = exc
        Exception.__init__(self, "Error loading user code: " + str(exc))


# Our jobs:
#
# 1. Assemble a virtual filesystem corresponding to an app's source code
# 2. (for Python 3 back-compatibility:) Intercept requests to import top-level modules in the main app,
#    and instead load the module in its main-app package, then alias it into sys.modules
#
class SimpleLoader(object):
    def __init__(self, module, real_name=None):
        self._module = module
        self._real_name = real_name

    def load_module(self, name):
        #print("Loading: " + name)
        if name in sys.modules:
            return sys.modules[name]

        real_name = self._real_name or name

        mod = self._module.get("module_object") or imp.new_module(real_name)
        sys.modules[real_name] = mod
        # Grungy horrid double-loading hack for Python 3
        if name != real_name:
            sys.modules[name] = mod

        if self._module.get('is_package'):
            mod.__path__ = []

        if 'code' in self._module:
            try:
                do_exec(compile(self._module['code'], real_name.replace(".", "/") + '.py', 'exec'), mod.__dict__)
            except Exception as e:
                raise ErrorLoadingUserCode(e)

        # Belt and braces? (honestly not sure about this -M)
        sys.modules[name] = mod

        return mod


class AppModuleFinder(object):

    def __init__(self):
        self._app = None
        self._modules = {}
        self._main_package = ''

    def _load_mods(self, app_spec, prefix=''):
        for module in app_spec.get('modules', []) + app_spec.get('server_modules', []):
            self._modules[prefix + module['name']] = module

        for form in app_spec.get('forms', []):
            self._modules[prefix + form['class_name']] = form

            if app_spec.get('runtime_options', {}).get('version', 0) < 2:
                template_mod = sys.modules['anvil']
            else:
                modname = prefix + form['class_name']
                if not form.get('is_package'):
                    modname = ".".join(modname.split(".")[:-1])
                modname += "._anvil_designer"
                if modname in self._modules:
                    template_mod = self._modules[modname]['module_object']
                else:
                    template_mod = imp.new_module(modname)
                    self._modules[modname] = {'module_object': template_mod}

            leaf_name = form['class_name'].split(".")[-1]
            setattr(template_mod, leaf_name+"Template", _form_templating.mk_template_class(form))


    def set_app(self, app_spec):
        self._app = app_spec
        self._main_package = app_spec.get('package_name', 'main_app_package')
        _form_templating.packages_by_app_id[''] = self._main_package

        deps = self._app.get('dependency_code', {})
        for dep_id in deps:
            dep = deps[dep_id]
            if 'package_name' in dep:
                _form_templating.packages_by_app_id[dep_id] = dep['package_name']
                self._modules[dep['package_name']] = {'is_package': True}
                self._load_mods(dep, dep['package_name']+".")

        self._modules[self._main_package] = {'is_package': True}
        self._load_mods(self._app, self._main_package+".")

    def get_main_package(self):
        return self._main_package

    def find_module(self, name, path=None):
        #print("Finding: " + name)
        #print(list(self._modules))
        mod = self._modules.get(name)
        if mod is not None:
            return SimpleLoader(mod)

        # Hack: Intercept top-level import requests

        #print("Initial miss on %s, trying %s" % (name, self._main_package+"."+name))
        mod = self._modules.get(self._main_package+"."+name)
        if mod is not None:
            return SimpleLoader(mod, self._main_package+"."+name)

    def app_is_loaded(self):
        return self._app is not None


module_finder = AppModuleFinder()
sys.meta_path.append(module_finder)

modules_to_import = []

def load_app(app):
    global modules_to_import

    module_finder.set_app(app)

    app_package = module_finder.get_main_package()
    modules_to_import = [app_package + "." + m['name'] for m in app.get("server_modules", [])]

    deps = app.get('dependency_code', {})
    for dep_app in deps.values():
        if "package_name" in dep_app:
            modules_to_import += [dep_app["package_name"] + "." + m['name'] for m in dep_app.get("server_modules", [])]

    # We have our app now. Anyone who's waiting should go ahead and execute.
    _serialise.release_reqresps()


def load_app_modules():
    """Call from _threaded_server when the environment is ready to import all server modules in this app"""
    try:
        for n in modules_to_import:
            importlib.import_module(n)
    except ErrorLoadingUserCode as e:
        raise e.exc

repl_scopes = {}


def handle_incoming_call(msg, send_to_host):
    if msg['type'].startswith("LAUNCH_BACKGROUND"):
        anvil.server.task_state = {}

    if not module_finder.app_is_loaded():
        if msg.get('app'):
            load_app(msg['app'])
        else:
            send_to_host({"type": "GET_APP", "id": _threaded_server.gen_id(), "originating-call": msg['id'],
                          "app-id": msg["app-id"], "app-version": msg["app-version"]})

    # This part happens out here because uplinks can't do REPLs:
    run_fn = None
    if msg['type'] == "LAUNCH_REPL":
        repl_scopes[msg['id']] = {}

        def run_fn():
            send_to_host({'output': "Application loaded\n", 'id': msg['id']})
            raise _threaded_server.SendNoResponse

    elif msg['type'] == "REPL_COMMAND":
        scope = repl_scopes.get(msg['repl'], {})

        def run_fn():
            cobj = compile(msg['command'], "<input>", "single")
            do_exec(cobj, scope)

    elif msg['type'] == "TERMINATE_REPL":
        repl_scopes.pop(msg['repl'], None)
        send_to_host({"id": msg['repl'], "response": None})
        send_to_host({"id": msg['id'], "response": None})
        return

    try:
        _threaded_server.IncomingRequest(msg, load_app_modules,
                                         run_fn=run_fn,
                                         dump_task_state=(msg['type'].startswith("LAUNCH_BACKGROUND")))
    except:
        send_to_host(_server._report_exception(msg['id']))



from setuptools import setup,find_packages
setup(
    name="anvil-app-server",
    version="1.6.1",
    packages=find_packages(),
    install_requires=["pychrome", "anvil-uplink==0.3.39", "progressbar2", "wheel", "psutil", "pgcli"],

    # Include the App Server JAR directly in this debug build of the package. That way the release version won't be downloaded on first run.
    package_data={
        "anvil_app_server": [
            "TodoList.zip",
            "Default.zip",
            "Blank.zip"
        ] 
    },

    author="Anvil",
    author_email="contact@anvil.works",
    description="A standalone server for Anvil full-stack Python web apps",
    long_description="""
## A standalone server for Anvil full-stack Python web apps.

This package includes several assets, some not in Python:
 - The "anvil-standalone-runtime" script for launching the runtime
 - The "downlink" Python code for the server-side parts of apps
 - The core Anvil server (requires Java to launch)
 - The client-side Anvil runtime (Javascript, using the Skulpt Python-to-JS compiler)

Head over to [the Github Repository](https://github.com/anvil-works/anvil-runtime) for usage instructions, and to learn more.
    """,
    long_description_content_type="text/markdown",
    keywords="anvil web apps standalone browser Python",
    url="https://anvil.works",
    license="GNU Affero General Public License v3",
    project_urls={
        "Source Code": "https://github.com/anvil-works/anvil-runtime"
    },
    classifiers=[
        "License :: OSI Approved :: GNU Affero General Public License v3"
    ],
    entry_points = {
        "console_scripts": [
        "anvil-app-server=anvil_app_server:launch",
        # For compat with instructions given in preview
        "anvil-standalone-runtime=anvil_app_server:launch",
        "psql-anvil-app-server=anvil_app_server:psql",
        "create-anvil-app=anvil_app_server:create_app"
        ]
    })



import anvil.server
import time

# Hack: Force ourselves into the top-level package, even
# if we were loaded into a runtime-v1 per-app Anvil package
__package__ = "anvil.tables"
__name__ = "anvil.tables"

class AppTables:
	cache = None

	def __getattr__(self, name):
		if AppTables.cache is None:
			AppTables.cache = anvil.server.call("anvil.private.tables.get_app_tables")

		tbl = AppTables.cache.get(name)
		if tbl is not None:
			return tbl

		raise AttributeError("No such app table: '%s'" % name)

	def __setattr__(self, name, val):
		raise Exception("app_tables is read-only")


#!defModuleAttr(anvil.tables)!1: 
# {
# 	name: "app_tables",
#	type: "any",
# 	anvil$helpLink: "/docs/data-tables/data-tables-in-code",
# 	$doc: "Access Table objects from the datatables services. You can access a Table object with dot notation e.g. `app_tables.my_table`. To access a table with strings use `getattr(app_tables, 'my_table')`. If no table is present an AttributeError will be thrown."
# }
#
app_tables = AppTables()


#!defClass(anvil.tables,TableError,__builtins__..Exception)!:
class TableError(anvil.server.AnvilWrappedError):
	pass


#!defClass(anvil.tables,TransactionConflict,anvil.tables.TableError)!:
class TransactionConflict(TableError):
	pass


#!defClass(anvil.tables,#QuotaExceededError,anvil.tables.TableError)!:
class QuotaExceededError(TableError):
	pass


anvil.server._register_exception_type("anvil.tables.TransactionConflict", TransactionConflict)
anvil.server._register_exception_type("anvil.tables.TableError", TableError)
anvil.server._register_exception_type("anvil.tables.QuotaExceededError", QuotaExceededError)


class Transaction:
	def __init__(self):
		self._aborting = False

	#!defMethod(anvil.tables.Transaction instance)!2: "Begin the transaction" ["__enter__"]
	def __enter__(self):
		anvil.server.call("anvil.private.tables.open_transaction")
		return self

	#!defMethod(_)!2: "End the transaction" ["__exit__"]
	def __exit__(self, e_type, e_val, tb):
		anvil.server.call("anvil.private.tables.close_transaction", self._aborting or e_val is not None)

	#!defMethod(_)!2: "Abort this transaction. When it ends, all write operations performed during it will be cancelled" ["abort"]
	def abort(self):
		self._aborting = True
#!defClass(anvil.tables,%Transaction)!:


#!defFunction(anvil.tables,%,function,server_function)!2: 
# {
#	$doc: "When applied to a function (as a decorator), the whole function will run in a data tables transaction. If it conflicts with another transaction, it will retry up to five times.",
# anvil$helpLink: "/docs/data-tables/transactions"
#  } ["in_transaction"]
def in_transaction(f):
	def new_f(*args, **kwargs):
		n = 0
		while True:
			try:
				with Transaction():
					return f(*args, **kwargs)
			except TransactionConflict:
                # lazy load random incase we make random.js a slow path on the client
				import random
				n += 1
				if n == 18:
					raise
				#print(f"RETRYING TXN {n}")
				# Max total sleep time is a little under 150 seconds (avg 75), so server calls will timeout before this finishes usually.
				sleep_amt = random.random() * (1.5**n) * 0.05
				try:
					time.sleep(sleep_amt)
				except:
					anvil.server.call("anvil.private._sleep", sleep_amt)

	try:
		reregister = f._anvil_reregister
	except AttributeError:
		pass
	else:
		reregister(new_f)

	new_f.__name__ = f.__name__

	return new_f


#!defFunction(anvil.tables,_,column_name,ascending=)!2: "Sort the results of this table search by a particular column. Default to ascending order." ["order_by"]
@anvil.server.portable_class
class order_by(object):
	def __init__(self, column_name, ascending=True):
		self.column_name = column_name
		self.ascending = ascending

@anvil.server.portable_class
class _page_size(object):
	def __init__(self, rows):
		self.rows = rows


#!defFunction(anvil.tables,%,[via_host=],[via_port=])!2: "Get a Postgres connection string for accessing this app's Data Tables via SQL.\n\nThe returned string includes temporary login credentials and sets the search path to a schema representing this app's Data Table environment.\n\nYou can override the host and port for the database connection to connect via a secure tunnel.\n\n(Available on the Dedicated Plan only.)" ["get_connection_string"]
def get_connection_string(via_host=None, via_port=None):
	return anvil.server.call("anvil.private.get_direct_postgres_connection_string", via_host=via_host, via_port=via_port)




#!defMethod(table row, **column_values)!2: "Add a row to the data table. Use keyword arguments to specify column values." ["add_row"]
#!defMethod(client readable view)!2: "Return a view on the table that can be read by client code. Use keyword arguments to specify view restrictions" ["client_readable"]
#!defMethod(client writable view)!2: "Return a view on the table that can be written by client code. Use keyword arguments to specify view restrictions. This does not give the client write access to other tables referred to by the table." ["client_writable"]
#!defMethod(client writable view)!2: "Return a view on this table that can be written by client code. Use keyword arguments to specify view restrictions." ["client_writable_cascade"]
#!defMethod(_)!2: "Delete all the rows from the data table" ["delete_all_rows"]
#!defMethod(_)!2: "Get a single matching row from the data table whose columns match the keyword arguments. Returns None if no matching row exists, and raises an exception if more than one row matches.\n\nEg: app_tables.table_1.get(name='John Smith')" ["get"]
#!defMethod(row,id)!2: "Get the matching row from this data table, by its unique ID" ["get_by_id"]
#!defMethod(bool,row)!2: "Returns true if the table (or view) contains the provided row." ["has_row"]
#!defMethod(list of dicts)!2: "Get the spec for the table as a list of dicts. Each dict contains the name and type of a column." ["list_columns"]
#!defMethod(Row or None)!2: "Get rows from a data table. If you specify keyword arguments, you will retrieve only rows whose columns match those values.\n\nEg: app_tables.table_1.search(name='John Smith')" ["search"]
#!defMethod(Media object)!2: "Get the table in CSV format. Returns a downloadable Media object; use its url property." ["to_csv"]
#!defClassNoConstructor(anvil.tables,#Table)!1: "A table returned from app_tables"

#!defMethod(Media object)!2: "Get the results of the SearchIterator in CSV format. Returns a downloadable Media object; use its url property." ["to_csv"]
#!defClassNoConstructor(anvil.tables,#SearchIterator)!1: "An iterator of table rows returned from a search()";


#!defMethod(_)!2: "Delete the row from its data table" ["delete"]
#!defMethod(id)!2: "Get the unique ID of the table row" ["get_id"]
#!defMethod(_,**column_values)!2: "update the data for multiple columns" ["update"]
#!defClassNoConstructor(anvil.tables,#Row)!1: "A table row";



import anvil.server
import sys

string_type = str
if sys.version_info < (3,):
    string_type = basestring

#!defFunction(anvil.google.mail,%,[to=],[subject=None],[text=None],[html=None],[cc=],[bcc=],[from_address=None],[draft=False])!2: "Send an email via GMail. 'to', 'cc' and 'bcc' may be strings (email addresses) or lists of strings (multiple addresses). At least one of 'text' and 'html' need to be provided (both strings). Passing draft=True will create a draft message rather than sending it." ["send"]
def send(to=[], subject=None, text=None, html=None, cc=[], bcc=[], from_address=None, draft=False):
    if not anvil.is_server_side():
        raise Exception("Only server modules can send email")

    def to_list(val):
        return [val] if isinstance(val, string_type) else list(val)

    to = to_list(to)
    cc = to_list(cc)
    bcc = to_list(bcc)

    if from_address is not None and not isinstance(from_address, string_type):
        raise Exception("from_address must be a string (eg 'John Smith <jsmith@example.com>'")

    if len(to + cc + bcc) == 0:
        raise Exception("You must specify at least one recipient (to, cc or bcc)")

    if text is None and html is None:
        raise Exception("You must supply a message body (text or html)")

    return anvil.server.call("anvil.private.google.mail.send", from_address=from_address, to=to, cc=cc, bcc=bcc, subject=subject, text=text, html=html, draft=draft)

